/* to get sys_errlist declared */
#define _BSD_SOURCE

#include <sys/socket.h>
#include <net/if.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netpacket/packet.h>
#include <unistd.h>
#include <time.h>
#include <sys/ioctl.h>
#include <resolv.h>
#include <errno.h>
#include <stdlib.h>
#include "uint16.h"
#include "fmt.h"
#include "buffer.h"
#include "ip4.h"
#include "ip6.h"
#include "byte.h"
#include "str.h"
#include "errmsg.h"
#include <sys/time.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>

#ifdef __dietlibc__
#include <write12.h>
#else
static void __write2(const char* s) {
  write(2,s,strlen(s));
}
#endif

#ifndef __dietlibc__
#define __attribute__(x)
#endif

int opte=0,optv=0,optn=0;

#define HOLES 20

struct hole {
  uint32 first,last;
};

unsigned long long connections;

struct connection {
  char srcip[16];
  char dstip[16];
  struct timeval tv,select;
  uint16 srcport;
  uint16 dstport;
  uint32 ofs;	/* inside the tcp window, last byte received */
  int trigger;
  int holes;
  int poisoned;
  struct hole h[HOLES];
  char* buf,* query;
  unsigned long n,a;
  time_t t;
  struct connection* buddy,* next;
}* root[65536];

unsigned int hash(const char* srcip,const char* dstip) {
  int i;
  unsigned int res;
  res=0;
  for (i=0; i<16; ++i)
    res+=(srcip[i] << i) + (dstip[15-i] << i);
  return ((res&0xffff0000) >> 16) ^ (res & 0xffff);
}

static struct connection* find_connection6(const char* srcip,const char* dstip,uint16 srcport,uint16 dstport) {
  struct connection* x=root[hash(srcip,dstip)];
  while (x) {
    if (!byte_diff(srcip,16,x->srcip) &&
	!byte_diff(dstip,16,x->dstip) &&
	srcport==x->srcport &&
	dstport==x->dstport) { return x; }
    x=x->next;
  }
  return 0;
}

static struct connection* create_connection6(const char* srcip,const char* dstip,uint16 srcport,uint16 dstport,struct timeval* tv) {
  struct connection* x=find_connection6(srcip,dstip,srcport,dstport);
  unsigned int h=hash(srcip,dstip);
  if (x) return x;	/* already have an entry.  probably syn resend...?! */
  x=malloc(sizeof(struct connection));
  if (!x) return 0;
  ++connections;
  byte_copy(x->srcip,16,srcip);
  byte_copy(x->dstip,16,dstip);
  x->srcport=srcport;
  x->dstport=dstport;
  x->buf=x->query=0;
  x->trigger=0;
  x->ofs=0;
  x->n=x->a=0;
  x->next=root[h];
  x->tv=*tv;
  return root[h]=x;
}

static void kill_connection(struct connection* x) {
  struct connection** y;
  if (!x) return;
  --connections;
  y=&root[hash(x->srcip,x->dstip)];
  while (*y!=x) y=&((*y)->next);
  if (*y==x) {
    *y=x->next;
    free(x->query);
    free(x->buf);
    free(x);
  }
}

#if 0
static void dumpconnections(void) {
  struct connection* x=root;
  buffer_puts(buffer_1,"known connections:\n");
  while (x) {
    char b1[IP4_FMT];
    char b2[IP4_FMT];
    char b3[FMT_ULONG+2];
    char b4[FMT_ULONG+2];
    b1[fmt_ip4(b1,x->srcip)]=0;	/* src ip */
    b2[fmt_ip4(b2,x->dstip)]=0;	/* dest ip */
    b3[fmt_ulong(b3,x->srcport)]=0;	/* src port */
    b4[fmt_ulong(b4,x->dstport)]=0;	/* dest port */
    buffer_putm(buffer_1,"  ",b1,":",b3," -> ",b2,":",b4,"\n");
    x=x->next;
  }
  buffer_putmflush(buffer_1,"\n");
}
#endif

unsigned long long packetloss;
unsigned long long resends;
unsigned long long poisoned;
unsigned long long totalpackets;
const int debug=0;

static void adddata(struct connection* x,const char* s,unsigned long ofs,unsigned long l) {
  if (l==0) return;	/* no data */
  if (ofs+l > ofs && !x->poisoned) {
    if (ofs > x->ofs) {
      ++packetloss;
      if (x->holes>=HOLES) {
	if (debug) printf("too many holes!\n");
	x->poisoned=1;
	++poisoned;
	return;
      }
      if (ofs-x->ofs < 0xfff) {
	x->h[x->holes].first=x->ofs;
	x->h[x->holes].last=ofs-1;
	++x->holes;
//	printf("new hole!\n");
	if (debug) {
	  int i;
	  printf("packet loss: %x %x (holes: ",ofs,x->ofs);
	  for (i=0; i<x->holes; ++i) {
	    printf("%s%x-%x",i?" ":"",x->h[i].first,x->h[i].last);
	  }
	  printf(")\n");
	}
//	if (ofs-x->ofs < 0xfff) {
	  x->ofs=ofs+l;
//	}
      }
      /* __write2("packet loss\n"); */
      return;
    }	/* missed packet in the middle */
    if (ofs+l < x->ofs) {
      int i,done;
      done=0;
      if (debug) printf("potential resend %x-%x\n",ofs,ofs+l);
      for (i=0; i<x->holes; ++i) {
	/* do we fill the hole completely? */
	if (ofs == x->h[i].first) {
	  if (ofs+l >= x->h[i].last) {
	    if (debug) printf(" -> filling hole %x-%x\n",x->h[i].first,x->h[i].last);
	    for (; i+1<x->holes; ++i)
	      x->h[i]=x->h[i+1];
	    --x->holes;
	    --packetloss;
	    done=1;
	    break;
	  } else {
	    /* shrink hole */
	    if (debug) printf(" -> shrinking hole %x-%x to %x-%x\n",x->h[i].first,x->h[i].last,ofs+l,x->h[i].last);
	    x->h[i].first=ofs+l;
	    done=1;
	  }
	} else if (ofs > x->h[i].first && (ofs < x->h[i].last || x->h[i].last<x->h[i].first)) {
	  /* start of new data lies within known hole; shrink hole */
	  if (ofs+l < x->h[i].last || x->h[i].last<x->h[i].first) {
	    if (debug) printf(" -> split hole %x-%x into %x-%x + %x-%x\n",
		   x->h[i].first,x->h[i].last,
		   x->h[i].first,ofs-1,
		   ofs+l,x->h[i].last);
	    ++packetloss;
	    /* split hole :-( */
	    if (x->holes >= HOLES) {
	      if (debug) printf("too many holes!\n");
	      x->poisoned=1;
	      ++poisoned;
	      return;
	    }
	    x->h[x->holes].first=ofs+l;
	    x->h[x->holes].last=x->h[i].last;
	    x->h[i].last=ofs-1;
	    ++x->holes;
	    done=1;
	  } else {
	    if (debug) printf(" -> filling hole (case 2) %x-%x\n",x->h[i].first,x->h[i].last);
	    --packetloss;
	    for (; i+1<x->holes; ++i)
	      x->h[i]=x->h[i+1];
	    --x->holes;
	    done=1;
	  }
	}
      }
      if (!done) {
	++resends;
	if (debug) printf(" -> resend!\n");
//	if (optv) __write2("resend\n");
      }
      return;
    }	/* already saw this one */
  }

  if (ofs < x->ofs) {
    uint32 rl=ofs-x->ofs;	/* the length of the overlapping segment */
    if (rl>l) return;		/* should not happen */
    s+=rl;
    l-=rl;
  }
  if (x->n+l > x->a) {
    char* tmp;
    x->a += l + 4096;
    tmp=realloc(x->buf,x->a);
    if (!tmp) return;		/* out of memory */
    x->buf=tmp;
  }
  byte_copy(x->buf+x->n,l,s);
  x->n+=l;
  x->ofs+=l;
}

void mysql_parse(struct connection* x,struct timeval* tv) {
  char* c=x->buf;
  uint32 plen;
  for (c=x->buf; (c >= x->buf) && (c+4 < x->buf+x->n); ) {
    plen=uint32_read(c)&0xffffff;
    if (plen<10 || c[4]!=3) {	/* vermutlich kein SELECT */
      c+=plen+4;
//      printf("plen: %d\n",plen);
      continue;
    }
    if (c+plen > x->buf+x->n) return;
    free(x->query);
    x->query=malloc(plen);
    if (!x->query) return;
    byte_copy(x->query,plen,c+5);
    x->query[plen-1]=0;
    {
      struct connection* y=find_connection6(x->dstip,x->srcip,x->dstport,x->srcport);
      if (y) { free(y->query); y->query=0; y->trigger=1; y->select=*tv; y->query=x->query; x->query=0; }
    }
    {
      unsigned int remove=c+5+plen-x->buf;
      if (x->n>=remove) {
	byte_copyr(x->buf,x->n-remove,x->buf+remove);
	x->n-=remove;
      } else {
	fprintf(stderr,"x->n is %lu, remove is %lu\n",x->n,remove);
      }
    }
    break;
  }
}

void http_parse(struct connection* x,struct timeval* tv) {
  char* c=x->buf;
  uint32 plen;
  char* host=0;
  char* req=0;
  if (x->n < 5) return;
  if (byte_equal(x->buf,4,"GET ") ||
      byte_equal(x->buf,5,"POST ") ||
      byte_equal(x->buf,5,"HEAD ")) {
    req=x->buf+4;
    if (*req==' ') ++req;
    for (plen=0; req+plen<x->buf+x->n; ++plen)
      if (req[plen]==' ') break;

    for (c=x->buf; c<x->buf+x->n; ++c) {
      if (!host && str_start(c,"\nHost: ")) {
	char* tmp;
	for (tmp=c+7; tmp<x->buf+x->n; ++tmp) {
	  if (*tmp=='\r' || *tmp=='\n') {
	    host=alloca(tmp-c);
	    byte_copy(host,tmp-c-7,c+7);
	    host[tmp-c-7]=0;
	    plen+=tmp-c-6;
	    break;
	  }
	}
      }
      if (byte_equal(c,4,"\r\n\r\n")) {
	int i;
	free(x->query);
	x->query=malloc(plen+20);
	for (i=0; x->buf[i]!=' '; ++i)
	  x->query[i]=x->buf[i];
	if (host) {
	  i+=fmt_str(x->query+i," http://");
	  i+=fmt_str(x->query+i,host);
	} else {
	  x->query[i]=' ';
	  ++i;
	}
	while (*req != ' ' && *req != '\r' && *req != '\n') {
	  x->query[i]=*req;
	  ++i;
	  ++req;
	}
	x->query[i]=0;
	{
	  struct connection* y=find_connection6(x->dstip,x->srcip,x->dstport,x->srcport);
	  if (y) { y->trigger=1; y->select=*tv; y->query=x->query; x->query=0; }
	}
	{
	  unsigned int remove=c+4-x->buf;
	  if (x->n>=remove) {
	    byte_copyr(x->buf,x->n-remove,x->buf+remove);
	    x->n-=remove;
	  } else {
	    fprintf(stderr,"x->n is %lu, remove is %lu\n",x->n,remove);
	  }
	}
      }
    }
  }
}


static void paket(const unsigned char* ip,const unsigned char* tcp,const unsigned char* payload,const unsigned char* max,struct timeval* tv,int v6) {
  unsigned char srcip[16];
  unsigned char dstip[16];
  uint16 srcport=uint16_read_big(tcp);
  uint16 dstport=uint16_read_big(tcp+2);
  enum { NONE, MYSQL, HTTP } mode=NONE;

  if (v6) {
    byte_copy(srcip,16,ip+8);
    byte_copy(dstip,16,ip+8+16);
  } else {
    byte_copy(srcip,12,V4mappedprefix);
    byte_copy(srcip+12,4,ip+12);
    byte_copy(dstip,12,V4mappedprefix);
    byte_copy(dstip+12,4,ip+16);
  }

  if (srcport==3306 || dstport==3306)
    mode=MYSQL;
  else if (srcport==80 || dstport==80)
    mode=HTTP;
  else
    return;

  if ((tcp[13]&2)==2) {	/* syn but no ack */
    struct connection* x;
#if 0
    char b1[IP6_FMT];
    char b2[IP6_FMT];
    char b3[FMT_ULONG+2];
    char b4[FMT_ULONG+2];
    b1[fmt_ip6c(b1,srcip)]=0;	/* src ip */
    b2[fmt_ip6c(b2,dstip)]=0;	/* dest ip */
    b3[fmt_ulong(b3,srcport)]=0;	/* src port */
    b4[fmt_ulong(b4,dstport)]=0;	/* dest port */
    buffer_putmflush(buffer_1,"connect ",b1,":",b3," -> ",b2,":",b4,"\n");
#endif
    x=create_connection6(srcip,dstip,srcport,dstport,tv);
    if (x) {
      x->ofs=uint32_read_big(tcp+4)+1;
      x->t=tv->tv_sec;
    }
  } else if (tcp[13]&1) {	/* FIN */
#if 0
    char b1[IP6_FMT];
    char b2[IP6_FMT];
    char b3[FMT_ULONG+2];
    char b4[FMT_ULONG+2];
    b1[fmt_ip6c(b1,srcip)]=0;	/* src ip */
    b2[fmt_ip6c(b2,dstip)]=0;	/* dest ip */
    b3[fmt_ulong(b3,srcport)]=0;	/* src port */
    b4[fmt_ulong(b4,dstport)]=0;	/* dest port */
    buffer_putmflush(buffer_1,"FIN ",b1,":",b3," -> ",b2,":",b4,"\n");
#endif
    kill_connection(find_connection6(srcip,dstip,srcport,dstport));
  } else {
    struct connection* x=find_connection6(srcip,dstip,srcport,dstport);
    if (x && (max-payload>0)) {
      x->t=tv->tv_sec;
      adddata(x,payload,uint32_read_big(tcp+4),max-payload);
      if (mode==MYSQL)
	mysql_parse(x,tv);
      else if (mode==HTTP)
	http_parse(x,tv);
      if (x->trigger && x->query) {
	char buf1[2*FMT_ULONG+10];
	int i,j;
	unsigned long l;
	for (i=j=0; x->query[i]; ++i) {
	  if (x->query[i]<' ')
	    x->query[j]=' ';
	  else
	    x->query[j]=x->query[i];
	  if (j==0 || x->query[j]!=' ' || x->query[j-1]!=' ') ++j;
	}
	x->query[j]=0;
	l=(tv->tv_sec-x->select.tv_sec)*1000000 + (tv->tv_usec-x->select.tv_usec);
	i=fmt_ulong(buf1,l/1000000);
	buf1[i]='.'; ++i;
	i+=fmt_ulong0(buf1+i,l%1000000,6);
	buf1[i]=0;
	buffer_putmflush(buffer_1,buf1,"\t",x->query,"\n");
	x->trigger=0;
	{
	  struct connection* y=find_connection6(dstip,srcip,dstport,srcport);
	  if (y) { y->trigger=0; y->n=0; }
	}
	x->n=0;
      }
    } else {
//      __write2("data for unknown connection!\n");
//      dumpconnections();
    }
  }
}

static void age_connections() {
  int i;
  time_t now=time(0);
  struct connection** x;
  for (i=0; i<65536; ++i) {
    for (x=&root[i]; *x; ) {
      if (now-(*x)->t >= 60*10) {	/* let old connections die (after 10 minutes) */
	struct connection* tmp=*x;
	*x=tmp->next;
	free(tmp->buf);
	free(tmp->query);
	free(tmp);
	--connections;
      } else
	x=&(*x)->next;
    }
  }
}

char myhostname[100];

void stats() {
  char pl[FMT_ULONG];
  char rs[FMT_ULONG];
  char pk[FMT_ULONG];
  char ps[FMT_ULONG];
  char cn[FMT_ULONG];
  pl[fmt_ulonglong(pl,packetloss)]=0;
  rs[fmt_ulonglong(rs,resends)]=0;
  pk[fmt_ulonglong(pk,totalpackets)]=0;
  ps[fmt_ulonglong(ps,poisoned)]=0;
  cn[fmt_ulonglong(cn,connections)]=0;
  buffer_putmflush(buffer_2,myhostname,": ",cn," connections, ",pl," packets lost, ",rs," resends, ",ps," connections poisoned, ",pk," packets total.\n");
  packetloss=resends=poisoned=totalpackets=0;
  age_connections();
}

int main(int argc,char* argv[]) {
  int sock,len;
  unsigned char buf[16*1024];
  struct sockaddr_ll sl,from;
  socklen_t fromlen=sizeof from;
  const char* interface="eth0";
  struct ifreq ifr;
  buffer fin;
  char* filename=0;
  int pcap_swap;
  struct timeval last;
  struct {
    uint32_t magic;
    uint16_t version_major,version_minor;
    int32_t thiszone;		/* gmt to local correction */
    uint32_t sigfigs, snaplen, linktype;
  } pcap_header;

  gethostname(myhostname,100);
  myhostname[99]=0;
  errmsg_iam("sspeed");

  for (;;) {
    int i;
    if ((i=getopt(argc,argv,"i:evnr:"))==-1) break;
    switch (i) {
    case 0:
      die(1,"non-option \"",argv[optind],"\"!");
    case 'i': interface=optarg; break;
    case 'e': opte=1; break;
    case 'v': optv=1; break;
    case 'n': optn=1; break;
    case 'r': filename=optarg; break;
    default:
      die(0,"usage: sspeed [-nev] [-i ifname] [-r dumpfile]");
    }
  }
  if (filename) {
    char bufbuf[8192];
    sock=open(filename,O_RDONLY);
    if (sock==-1) diesys(1,"open \"",filename,"\"");
    if (read(sock,&pcap_header,sizeof(pcap_header)) != sizeof(pcap_header))
      die(1,"\"",filename,"\": file truncated.");
    if (pcap_header.magic != 0xa1b2c3d4)
      die(1,"invalid pcap capture file magic!");
    buffer_init(&fin,read,sock,bufbuf,sizeof bufbuf);
    pcap_swap=(pcap_header.version_major<2 || (pcap_header.version_major==2 && pcap_header.version_minor<=3));
    if (pcap_header.linktype != 1)	/* DLT_EN10MB */
      die(1,"non-ethernet dump!");
    ifr.ifr_hwaddr.sa_family=ARPHRD_ETHER;
  } else {
    buffer_puts(buffer_2,"using interface ");
    buffer_puts(buffer_2,interface);
    buffer_putnlflush(buffer_2);
    sock=socket(PF_PACKET,SOCK_RAW,ETH_P_ALL);
    if (sock==-1) diesys(1,"socket");
    sl.sll_family=AF_PACKET;
    sl.sll_protocol=htons(ETH_P_ALL);
    sl.sll_ifindex=if_nametoindex(interface);
    if (bind(sock,(struct sockaddr*)&sl,sizeof sl)==-1) diesys(1,"bind");
    {
      strncpy(ifr.ifr_name,interface,IFNAMSIZ);
      if (ioctl(sock,SIOCGIFHWADDR, &ifr)==-1) diesys(1,"SIOCGIFINDEX");
      switch (ifr.ifr_hwaddr.sa_family) {
      case ARPHRD_ETHER:
      case ARPHRD_LOOPBACK:
      case ARPHRD_IEEE80211: // WLAN
      case ARPHRD_PPP:
	break;
      default:
	buffer_puts(buffer_2,"unknown layer 2: ");
	buffer_putulong(buffer_2,ifr.ifr_hwaddr.sa_family);
	buffer_putnlflush(buffer_2);
      }
    }
  }
  for (;;) {
    struct timeval tv;
    unsigned char* max;

    if (filename) {
      struct pcap_timeval {
	int32_t tv_sec;           /* seconds */
	int32_t tv_usec;          /* microseconds */
      };

      struct pcap_sf_pkthdr {
	struct pcap_timeval ts;     /* time stamp */
	uint32_t caplen;         /* length of portion present */
	uint32_t len;            /* length this packet (off wire) */
      };

      struct pcap_sf_pkthdr p;
      if ((len=buffer_get(&fin,(char*)&p,sizeof(p))!=sizeof(p))) {
	stats();
	die(1,"short read");
      }
      tv.tv_sec=p.ts.tv_sec;
      tv.tv_usec=p.ts.tv_usec;
//      printf("len %d caplen %d\n",p.len,p.caplen);
      if ((len=buffer_get(&fin,buf,p.caplen)!=p.caplen)) {
	stats();
	die(1,"short read");
      }
      len=p.len;
    } else {
      if ((len=recvfrom(sock,buf,sizeof buf,0,(struct sockaddr*)&from,&fromlen))<=0) break;
      if (len<14+20+20) continue;
      if (ifr.ifr_hwaddr.sa_family==ARPHRD_LOOPBACK && from.sll_pkttype!=PACKET_OUTGOING)
	/* Linux gives every packet twice on loopback.  Only look at
	* outgoing packets */
	continue;
      ioctl(sock,SIOCGSTAMP,&tv);
    }

#if 0
    if (tv.tv_sec == 0x42f73b41 && tv.tv_usec == 0x000cc270)
      printf("%x:%08x      \r",tv.tv_sec,tv.tv_usec);
#endif

    max=buf+len;

    switch (ifr.ifr_hwaddr.sa_family) {
    case ARPHRD_ETHER:
    case ARPHRD_LOOPBACK:
      if (uint16_read_big(buf+12)==0x0800) /* IP */ {
	unsigned char* ip=buf+14;
	unsigned char* tcp;
	unsigned char* payload;
	if ((ip[0]&0xf0) != 0x40) continue;	/* not IPv4 */
	if ((ip[9] != 6)) continue;		/* not TCP */
	tcp=ip+(ip[0]&0xf)*4;
	if (tcp+20>max) continue;
	payload=tcp+(tcp[12]>>4)*4;
	if (payload>max) continue;
	paket(ip,tcp,payload,max,&tv,0);
      } else if (uint16_read_big(buf+12)==0x86dd) /* IPv6 */ {
	unsigned char* ip=buf+14;
	unsigned char* tcp;
	unsigned char* payload;
	if ((ip[0]&0xf0) != 0x60) continue;	/* not IPv4 */
	tcp=ip+8+2*16;
	if ((ip[6] != 6)) continue;		/* not TCP */
	if (tcp+20>max) continue;
	payload=tcp+(tcp[12]>>4)*4;
	if (payload>max) continue;
	paket(ip,tcp,payload,max,&tv,1);
      }
      break;
    }
    ++totalpackets;
    if (totalpackets >= 10000) last.tv_sec=0;
    if (last.tv_sec != tv.tv_sec) {
      struct tm a;
      struct tm b;
      if (totalpackets == 1) continue;
      gmtime_r(&tv.tv_sec,&a);
      gmtime_r(&last.tv_sec,&b);
//      printf("%02d:%02d vs %02d:%02d\n",a.tm_hour,a.tm_min,b.tm_hour,b.tm_min);
      if ((a.tm_min != b.tm_min) ||
	  (a.tm_hour != b.tm_hour)) {
	stats();
	last=tv;
      }
    }
  }
  diesys(1,"recv");
  return 0;
}
